# -*- coding: utf-8 -*-
"""
Modular Canvas Rendering
========================

Module for visualizing model objects in grid cells.

"""
from collections import defaultdict
from mesa.visualization.ModularVisualization import VisualizationElement


class CanvasGrid(VisualizationElement):
    """ A CanvasGrid object uses a user-provided portrayal method to generate a
    portrayal for each object. A portrayal is a JSON-ready dictionary which
    tells the relevant JavaScript code (GridDraw.js) where to draw what shape.

    The render method returns a dictionary, keyed on layers, with values as
    lists of portrayals to draw. Portrayals themselves are generated by the
    user-provided portrayal_method, which accepts an object as an input and
    produces a portrayal of it.

    A portrayal as a dictionary with the following structure:
        "x", "y": Coordinates for the cell in which the object is placed.
        "Shape": Can be either "circle", "rect" or "arrowHead"
            For Circles:
                "r": The radius, defined as a fraction of cell size. r=1 will
                     fill the entire cell.
            For Rectangles:
                "w", "h": The width and height of the rectangle, which are in
                          fractions of cell width and height.
            For arrowHead:
            "scale": Proportion scaling as a fraction of cell size.
            "heading_x": represents x direction unit vector.
            "heading_y": represents y direction unit vector.
        "Color": The color to draw the shape in; needs to be a valid HTML
                 color, e.g."Red" or "#AA08F8"
        "Filled": either "true" or "false", and determines whether the shape is
                  filled or not.
        "Layer": Layer number of 0 or above; higher-numbered layers are drawn
                 above lower-numbered layers.
        "text": The text to be inscribed inside the Shape. Normally useful for
                showing the unique_id of the agent.
        "text_color": The color to draw the inscribed text. Should be given in
                      conjunction of "text" property.


    Attributes:
        portrayal_method: Function which generates portrayals from objects, as
                          described above.
        grid_height, grid_width: Size of the grid to visualize, in cells.
        canvas_height, canvas_width: Size, in pixels, of the grid visualization
                                     to draw on the client.
        template: "canvas_module.html" stores the module's HTML template.

    """
    package_includes = ["GridDraw.js", "CanvasModule.js", "InteractionHandler.js"]

    def __init__(self, portrayal_method, grid_width, grid_height,
                 canvas_width=500, canvas_height=500):
        """ Instantiate a new CanvasGrid.

        Args:
            portrayal_method: function to convert each object on the grid to
                              a portrayal, as described above.
            grid_width, grid_height: Size of the grid, in cells.
            canvas_height, canvas_width: Size of the canvas to draw in the
                                         client, in pixels. (default: 500x500)

        """
        self.portrayal_method = portrayal_method
        self.grid_width = grid_width
        self.grid_height = grid_height
        self.canvas_width = canvas_width
        self.canvas_height = canvas_height

        new_element = ("new CanvasModule({}, {}, {}, {})"
            .format(self.canvas_width, self.canvas_height,
                self.grid_width, self.grid_height))

        self.js_code = "elements.push(" + new_element + ");"

    def render(self, model):
        grid_state = defaultdict(list)
        for x in range(model.grid.width):
            for y in range(model.grid.height):
                cell_objects = model.grid.get_cell_list_contents([(x, y)])
                for obj in cell_objects:
                    portrayal = self.portrayal_method(obj)
                    if portrayal:
                        portrayal["x"] = x
                        portrayal["y"] = y
                        grid_state[portrayal["Layer"]].append(portrayal)

        return grid_state



# -*- coding: utf-8 -*-
"""
Mesa Agent-Based Modeling Framework

Core Objects: Model, and Agent.

"""
import datetime

from .model import Model
from .agent import Agent


__all__ = ["Model", "Agent"]

__title__ = 'mesa'
__version__ = '0.8.5'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright %s Project Mesa Team' % datetime.date.today().year



# -*- coding: utf-8 -*-
"""
The model class for Mesa framework.

Core Objects: Model

"""
import time
import random


class Model:
    """ Base class for models. """

    def __new__(cls, *args, **kwargs):
        """Create a new model object and instantiate its RNG automatically."""

        model = object.__new__(cls)  # This only works in Python 3.3 and above
        model._seed = time.time()
        if "seed" in kwargs and kwargs["seed"] is not None:
            model._seed = kwargs["seed"]
        model.random = random.Random(model._seed)
        return model

    def __init__(self):
        """ Create a new model. Overload this method with the actual code to
        start the model.

        Attributes:
            schedule: schedule object
            running: a bool indicating if the model should continue running

        """

        self.running = True
        self.schedule = None
        self.current_id = 0

    def run_model(self):
        """ Run the model until the end condition is reached. Overload as
        needed.

        """
        while self.running:
            self.step()

    def step(self):
        """ A single step. Fill in here. """
        pass

    def next_id(self):
        """ Return the next unique ID for agents, increment current_id"""
        self.current_id += 1
        return self.current_id

    def reset_randomizer(self, seed=None):
        """Reset the model random number generator.

        Args:
            seed: A new seed for the RNG; if None, reset using the current seed
        """

        if seed is None:
            seed = self._seed
        self.random.seed(seed)
        self._seed = seed



import sys
import os
import click
from subprocess import call

PROJECT_PATH = click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True)
COOKIECUTTER_DIR = 'mesa/cookiecutter-mesa'
SCRIPTS_DIR = os.path.dirname(os.path.abspath(__file__))
COOKIECUTTER_PATH = os.path.join(os.path.dirname(SCRIPTS_DIR),
                                 COOKIECUTTER_DIR)


@click.group()
def cli():
    'Manage Mesa projects'
    pass


@cli.command()
@click.argument('project', type=PROJECT_PATH, default='.')
def runserver(project):
    '''Run mesa project PROJECT

    PROJECT is the path to the directory containing `run.py`, or the current
    directory if not specified.
    '''
    sys.path.insert(0, project)
    os.chdir(project)

    with open("run.py") as f:
        code = compile(f.read(), "run.py", 'exec')
        exec(code, {}, {})


@click.command()
@click.option('--no-input', is_flag=True,
              help='Do not prompt user for custom mesa model input.')
def startproject(no_input):
    args = ['cookiecutter', COOKIECUTTER_PATH]
    if no_input:
        args.append('--no-input')
    call(args)


cli.add_command(runserver)
cli.add_command(startproject)



# -*- coding: utf-8 -*-
"""
Text Module
============

Module for drawing live-updating text.

"""
from mesa.visualization.ModularVisualization import VisualizationElement


class TextElement(VisualizationElement):
    package_includes = ["TextModule.js"]
    js_code = "elements.push(new TextModule());"



# -*- coding: utf-8 -*-
"""
Network Visualization Module
============

Module for rendering the network, using [sigma.js](http://sigmajs.org/) or [d3.js](https://d3js.org/) frameworks.

"""
from mesa.visualization.ModularVisualization import VisualizationElement


class NetworkModule(VisualizationElement):
    package_includes = []

    def __init__(self, portrayal_method, canvas_height=500, canvas_width=500, library='sigma'):
        library_types = ['sigma', 'd3']
        if library not in library_types:
            raise ValueError("Invalid javascript library type. Expected one of: %s" % library_types)

        NetworkModule.package_includes = ["NetworkModule_sigma.js", "sigma.min.js"] if library == 'sigma' else [
            "NetworkModule_d3.js", "d3.min.js"]

        self.portrayal_method = portrayal_method
        self.canvas_height = canvas_height
        self.canvas_width = canvas_width
        new_element = ("new NetworkModule({}, {})".
                       format(self.canvas_width, self.canvas_height))
        self.js_code = "elements.push(" + new_element + ");"

    def render(self, model):
        return self.portrayal_method(model.G)



# -*- coding: utf-8 -*-
"""
Batchrunner
===========

A single class to manage a batch run or parameter sweep of a given model.

"""
import copy
from itertools import product, count
import pandas as pd
from tqdm import tqdm
try:
    from pathos.multiprocessing import ProcessPool
except ImportError:
    pathos_support = False
else:
    pathos_support = True


class VariableParameterError(TypeError):
    MESSAGE = ('variable_parameters must map a name to a sequence of values. '
               'These parameters were given with non-sequence values: {}')

    def __init__(self, bad_names):
        self.bad_names = bad_names

    def __str__(self):
        return self.MESSAGE.format(self.bad_names)


class BatchRunner:
    """ This class is instantiated with a model class, and model parameters
    associated with one or more values. It is also instantiated with model and
    agent-level reporters, dictionaries mapping a variable name to a function
    which collects some data from the model or its agents at the end of the run
    and stores it.

    Note that by default, the reporters only collect data at the *end* of the
    run. To get step by step data, simply have a reporter store the model's
    entire DataCollector object.

    """
    def __init__(self, model_cls, variable_parameters=None,
                 fixed_parameters=None, iterations=1, max_steps=1000,
                 model_reporters=None, agent_reporters=None,
                 display_progress=True):
        """ Create a new BatchRunner for a given model with the given
        parameters.

        Args:
            model_cls: The class of model to batch-run.
            variable_parameters: Dictionary of parameters to lists of values.
                The model will be run with every combo of these paramters.
                For example, given variable_parameters of
                    {"param_1": range(5),
                     "param_2": [1, 5, 10]}
                models will be run with {param_1=1, param_2=1},
                    {param_1=2, param_2=1}, ..., {param_1=4, param_2=10}.
            fixed_parameters: Dictionary of parameters that stay same through
                all batch runs. For example, given fixed_parameters of
                    {"constant_parameter": 3},
                every instantiated model will be passed constant_parameter=3
                as a kwarg.
            iterations: The total number of times to run the model for each
                combination of parameters.
            max_steps: Upper limit of steps above which each run will be halted
                if it hasn't halted on its own.
            model_reporters: The dictionary of variables to collect on each run
                at the end, with variable names mapped to a function to collect
                them. For example:
                    {"agent_count": lambda m: m.schedule.get_agent_count()}
            agent_reporters: Like model_reporters, but each variable is now
                collected at the level of each agent present in the model at
                the end of the run.
            display_progress: Display progresss bar with time estimation?

        """
        self.model_cls = model_cls
        if variable_parameters is None:
            variable_parameters = {}
        self.variable_parameters = self._process_parameters(variable_parameters)
        self.fixed_parameters = fixed_parameters or {}
        self._include_fixed = len(self.fixed_parameters.keys()) > 0
        self.iterations = iterations
        self.max_steps = max_steps

        self.model_reporters = model_reporters
        self.agent_reporters = agent_reporters

        if self.model_reporters:
            self.model_vars = {}

        if self.agent_reporters:
            self.agent_vars = {}

        self.display_progress = display_progress

    def _process_parameters(self, params):
        params = copy.deepcopy(params)
        bad_names = []
        for name, values in params.items():
            if (isinstance(values, str) or not hasattr(values, "__iter__")):
                bad_names.append(name)
        if bad_names:
            raise VariableParameterError(bad_names)
        return params

    def _make_model_args(self):
        """Prepare all combinations of parameter values for `run_all`

        Returns:
            Tuple with the form:
            (total_iterations, all_kwargs, all_param_values)
        """
        total_iterations = self.iterations
        all_kwargs = []
        all_param_values = []

        if len(self.variable_parameters) > 0:
            param_names, param_ranges = zip(*self.variable_parameters.items())
            for param_range in param_ranges:
                total_iterations *= len(param_range)

            for param_values in product(*param_ranges):
                kwargs = dict(zip(param_names, param_values))
                kwargs.update(self.fixed_parameters)
                all_kwargs.append(kwargs)
                all_param_values.append(param_values)
        else:
            kwargs = self.fixed_parameters
            param_values = None
            all_kwargs = [kwargs]
            all_param_values = [None]

        return (total_iterations, all_kwargs, all_param_values)

    def run_all(self):
        """ Run the model at all parameter combinations and store results. """
        run_count = count()
        total_iterations, all_kwargs, all_param_values = self._make_model_args()

        with tqdm(total_iterations, disable=not self.display_progress) as pbar:
            for i, kwargs in enumerate(all_kwargs):
                param_values = all_param_values[i]
                for _ in range(self.iterations):
                    self.run_iteration(kwargs, param_values, next(run_count))
                    pbar.update()

    def run_iteration(self, kwargs, param_values, run_count):
        kwargscopy = copy.deepcopy(kwargs)
        model = self.model_cls(**kwargscopy)
        self.run_model(model)

        # Collect and store results:
        if param_values is not None:
            model_key = param_values + (run_count,)
        else:
            model_key = (run_count,)

        if self.model_reporters:
            self.model_vars[model_key] = self.collect_model_vars(model)
        if self.agent_reporters:
            agent_vars = self.collect_agent_vars(model)
            for agent_id, reports in agent_vars.items():
                agent_key = model_key + (agent_id,)
                self.agent_vars[agent_key] = reports
        return (getattr(self, "model_vars", None), getattr(self, "agent_vars", None))

    def run_model(self, model):
        """ Run a model object to completion, or until reaching max steps.

        If your model runs in a non-standard way, this is the method to modify
        in your subclass.

        """
        while model.running and model.schedule.steps < self.max_steps:
            model.step()

    def collect_model_vars(self, model):
        """ Run reporters and collect model-level variables. """
        model_vars = {}
        for var, reporter in self.model_reporters.items():
            model_vars[var] = reporter(model)
        return model_vars

    def collect_agent_vars(self, model):
        """ Run reporters and collect agent-level variables. """
        agent_vars = {}
        for agent in model.schedule._agents.values():
            agent_record = {}
            for var, reporter in self.agent_reporters.items():
                agent_record[var] = getattr(agent, reporter)
            agent_vars[agent.unique_id] = agent_record
        return agent_vars

    def get_model_vars_dataframe(self):
        """ Generate a pandas DataFrame from the model-level variables
        collected.

        """
        return self._prepare_report_table(self.model_vars)

    def get_agent_vars_dataframe(self):
        """ Generate a pandas DataFrame from the agent-level variables
        collected.

        """
        return self._prepare_report_table(self.agent_vars,
                                          extra_cols=['AgentId'])

    def _prepare_report_table(self, vars_dict, extra_cols=None):
        """
        Creates a dataframe from collected records and sorts it using 'Run'
        column as a key.
        """
        extra_cols = ['Run'] + (extra_cols or [])
        index_cols = list(self.variable_parameters.keys()) + extra_cols

        records = []
        for param_key, values in vars_dict.items():
            record = dict(zip(index_cols, param_key))
            record.update(values)
            records.append(record)

        df = pd.DataFrame(records)
        rest_cols = set(df.columns) - set(index_cols)
        ordered = df[index_cols + list(sorted(rest_cols))]
        ordered.sort_values(by='Run', inplace=True)
        if self._include_fixed:
            for param in self.fixed_parameters.keys():
                val = self.fixed_parameters[param]

                # avoid error when val is an iterable
                vallist = [val for i in range(ordered.shape[0])]
                ordered[param] = vallist
        return ordered


class MPSupport(Exception):
    def __str__(self):
        return ("BatchRunnerMP depends on pathos, which is either not "
               "installed, or the path can not be found. ")


class BatchRunnerMP(BatchRunner):
    """ Child class of BatchRunner, extended with multiprocessing support. """

    def __init__(self, model_cls, nr_processes=2, **kwargs):
        """ Create a new BatchRunnerMP for a given model with the given
        parameters.

        Args:
            model_cls: The class of model to batch-run.
            nr_processes: the number of separate processes the BatchRunner
                should start, all running in parallel.
            kwargs: the kwargs required for the parent BatchRunner class
        """
        if not pathos_support:
            raise MPSupport
        super().__init__(model_cls, **kwargs)
        self.pool = ProcessPool(nodes=nr_processes)

    def run_all(self):
        """
        Run the model at all parameter combinations and store results,
        overrides run_all from BatchRunner.
        """
        run_count = count()
        total_iterations, all_kwargs, all_param_values = self._make_model_args()

        # register the process pool and init a queue
        job_queue = []
        with tqdm(total_iterations, disable=not self.display_progress) as pbar:
            for i, kwargs in enumerate(all_kwargs):
                param_values = all_param_values[i]
                for _ in range(self.iterations):
                    # make a new process and add it to the queue
                    job_queue.append(self.pool.uimap(self.run_iteration,
                                                     (kwargs,),
                                                     (param_values,),
                                                     (next(run_count),)))
            # empty the queue
            results = []
            for task in job_queue:
                for model_vars, agent_vars in list(task):
                    results.append((model_vars, agent_vars))
                pbar.update()

            # store the results
            for model_vars, agent_vars in results:
                if self.model_reporters:
                    for model_key, model_val in model_vars.items():
                        self.model_vars[model_key] = model_val
                if self.agent_reporters:
                    for agent_key, reports in agent_vars.items():
                        self.agent_vars[agent_key] = reports




class UserSettableParameter:
    """ A class for providing options to a visualization for a given parameter.

        UserSettableParameter can be used instead of keyword arguments when specifying model parameters in an
        instance of a `ModularServer` so that the parameter can be adjusted in the UI without restarting the server.

        Validation of correctly-specified params happens on startup of a `ModularServer`. Each param is handled
        individually in the UI and sends callback events to the server when an option is updated. That option is then
        re-validated, in the `value.setter` property method to ensure input is correct from the UI to `reset_model`
        callback.

        Parameter types include:
            - 'number' - a simple numerical input
            - 'checkbox' - boolean checkbox
            - 'choice' - String-based dropdown input, for selecting choices within a model
            - 'slider' - A number-based slider input with settable increment
            - 'static_text' - A non-input textbox for displaying model info.

        Examples:

        # Simple number input
        number_option = UserSettableParameter('number', 'My Number', value=123)

        # Checkbox input
        boolean_option = UserSettableParameter('checkbox', 'My Boolean', value=True)

        # Choice input
        choice_option = UserSettableParameter('choice', 'My Choice', value='Default choice',
                                              choices=['Default Choice', 'Alternate Choice'])

        # Slider input
        slider_option = UserSettableParameter('slider', 'My Slider', value=123, min_value=10, max_value=200, step=0.1)

        # Static text
        static_text = UserSettableParameter('static_text', value="This is a descriptive textbox")
     """

    NUMBER = 'number'
    CHECKBOX = 'checkbox'
    CHOICE = 'choice'
    SLIDER = 'slider'
    STATIC_TEXT = 'static_text'

    TYPES = (NUMBER, CHECKBOX, CHOICE, SLIDER, STATIC_TEXT)

    _ERROR_MESSAGE = "Missing or malformed inputs for '{}' Option '{}'"

    def __init__(
        self, param_type=None, name='', value=None, min_value=None, max_value=None,
            step=1, choices=list(), description=None
    ):
        if param_type not in self.TYPES:
            raise ValueError("{} is not a valid Option type".format(param_type))
        self.param_type = param_type
        self.name = name
        self._value = value
        self.min_value = min_value
        self.max_value = max_value
        self.step = step
        self.choices = choices
        self.description = description

        # Validate option types to make sure values are supplied properly
        msg = self._ERROR_MESSAGE.format(self.param_type, name)
        valid = True

        if self.param_type == self.NUMBER:
            valid = not (self.value is None)

        elif self.param_type == self.SLIDER:
            valid = not (self.value is None or self.min_value is None or self.max_value is None)

        elif self.param_type == self.CHOICE:
            valid = not (self.value is None or len(self.choices) == 0)

        elif self.param_type == self.CHECKBOX:
            valid = isinstance(self.value, bool)

        elif self.param_type == self.STATIC_TEXT:
            valid = isinstance(self.value, str)

        if not valid:
            raise ValueError(msg)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        if self.param_type == self.SLIDER:
            if self._value < self.min_value:
                self._value = self.min_value
            elif self._value > self.max_value:
                self._value = self.max_value
        elif self.param_type == self.CHOICE:
            if self._value not in self.choices:
                print("Selected choice value not in available choices, selected first choice from 'choices' list")
                self._value = self.choices[0]

    @property
    def json(self):
        result = self.__dict__.copy()
        result['value'] = result.pop('_value')  # Return _value as value, value is the same
        return result



# -*- coding: utf-8 -*-
"""
The agent class for Mesa framework.

Core Objects: Agent

"""


class Agent:
    """ Base class for a model agent. """
    def __init__(self, unique_id, model):
        """ Create a new agent. """
        self.unique_id = unique_id
        self.model = model

    def step(self):
        """ A single step of the agent. """
        pass

    @property
    def random(self):
        return self.model.random
